 
/*
   Driver for GPIO pins
*/ 

#include <linux/module.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/uaccess.h>	/* for copy_from_user */


#include <mach/hardware.h>
#include <mach/gpio.h>

#define GPIO_AUTHOR        "Michael Schwarz"
#define GPIO_DESCRIPTION   "GPIO Driver for GNUBLIN"

#define PROCFS_MAX_SIZE         4

#define PROCFS_DIR  "gpio_dir"
#define PROCFS_IN   "gpio_in"
#define PROCFS_OUT  "gpio_out"

int gpio_out = 0;
int gpio_dir = 0;
int gpio_in  = 0;
static char procfs_buffer[PROCFS_MAX_SIZE];
static unsigned long procfs_buffer_size = 0;

int gpios[] = {GPIO_GPIO11, GPIO_GPIO14, GPIO_GPIO15, GPIO_SPI_MISO, GPIO_SPI_MOSI, GPIO_SPI_SCK, GPIO_I2C_SDA1, GPIO_I2C_SCL1, GPIO_PWM_DATA};
int gpio_cnt = 9;

int gpio_dir_write(struct file* file, const char* buffer, unsigned long count, void* data) {
 procfs_buffer_size = (count > PROCFS_MAX_SIZE ? PROCFS_MAX_SIZE : count);
 
 if(copy_from_user(procfs_buffer, buffer, procfs_buffer_size)) {
  return -EFAULT; 
 }
 
 gpio_dir = procfs_buffer[0] + (procfs_buffer[1] << 8);
 
 int i;
 for(i = 0; i < gpio_cnt; i++) {
  if(gpio_dir & (1 << i)) gpio_direction_output(gpios[i], 0); else gpio_direction_input(gpios[i]); 
 }

 
 return procfs_buffer_size;
}

int gpio_dir_read(char* buffer, char** buffer_location, off_t offset, int buffer_length, int* eof, void* data) {
 if(offset > 0) return 0;
 
 buffer[0] = gpio_dir & 0xff;
 buffer[1] = (gpio_dir >> 8) & 0xff;
 return 2;
}


int gpio_out_write(struct file* file, const char* buffer, unsigned long count, void* data) {
 procfs_buffer_size = (count > PROCFS_MAX_SIZE ? PROCFS_MAX_SIZE : count);
 
 if(copy_from_user(procfs_buffer, buffer, procfs_buffer_size)) {
  return -EFAULT; 
 }
 
 gpio_out = procfs_buffer[0];
 
 //printk("gpio-module: out %d%d%d\n", gpio_out & 1, (gpio_out & 2) > 0, (gpio_out & 4) > 0);
 
 int i;
 for(i = 0; i < gpio_cnt; i++) {
  if(gpio_out & (1 << i)) gpio_set_value(gpios[i], 1); else gpio_set_value(gpios[i], 0); 
 }
 
 return procfs_buffer_size;
}

int gpio_in_read(char* buffer, char** buffer_location, off_t offset, int buffer_length, int* eof, void* data) {
 if(offset > 0) return 0;
 
 gpio_in = 0;

 int i;
 for(i = 0; i < gpio_cnt; i++) {
  if(gpio_get_value(gpios[i])) gpio_in |= (1 << i); 
 }
 
 buffer[0] = gpio_in & 0xff;
 buffer[1] = (gpio_in >> 8) & 0xff;
 return 2; 
}


void create_procfs(char* name, void* fnc_read, void* fnc_write) {
 struct proc_dir_entry *proc_file;
 proc_file = create_proc_entry(name, 0644, NULL);

 if(proc_file == NULL) {
  remove_proc_entry(name, NULL);
  printk(KERN_ALERT "Error: Could not initialize /proc/%s\n", name);
  return;// -ENOMEM;
 }

 proc_file->read_proc  = fnc_read;
 proc_file->write_proc = fnc_write;
 proc_file->mode       = S_IFREG | S_IRUGO;
 proc_file->uid        = 0;
 proc_file->gid        = 0;
 proc_file->size       = 37;
 
}


int __init init_gpio(void) {
 printk("gpio-module: init\n");
    
 create_procfs(PROCFS_DIR, gpio_dir_read, gpio_dir_write);
 create_procfs(PROCFS_IN, gpio_in_read, NULL);
 create_procfs(PROCFS_OUT, NULL, gpio_out_write);
    
 return 0;
}

void __exit cleanup_gpio(void) {
 printk("gpio-module: cleanup\n");
   
 remove_proc_entry(PROCFS_DIR, NULL);
 remove_proc_entry(PROCFS_IN, NULL);
 remove_proc_entry(PROCFS_OUT, NULL);
}

module_init(init_gpio);
module_exit(cleanup_gpio);

MODULE_LICENSE("GPL");

MODULE_AUTHOR(GPIO_AUTHOR);  
MODULE_DESCRIPTION(GPIO_DESCRIPTION);    


